# Generated by Django 2.2.17 on 2021-01-04 16:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edx_proctoring', '0011_allow_multiple_attempts'),
    ]

    operations = [
        migrations.AddField(
            model_name='proctoredexamstudentattempt',
            name='time_remaining_seconds',
            field=models.IntegerField(null=True),
        ),
    ]
