{% extends 'practice_exam/entrance.html' %}
{% load i18n %}

{% block title %}
  {% blocktrans %}
    Proctoring onboarding exam
  {% endblocktrans %}
{% endblock %}

{% block body %}
  <p>
    {% blocktrans %}
      Why this is important to you:
    {% endblocktrans %}
  </p>
  <ul>
    <li>
      {% blocktrans %}
        Establish your identity with the proctoring system to take a proctored exam
      {% endblocktrans %}
    </li>
    <li>
      {% blocktrans %}
        Create your onboarding profile for faster access in the future
      {% endblocktrans %}
    </li>
    <li>
      {% blocktrans %}
        Practice taking a proctored test
      {% endblocktrans %}
    </li>
  </ul>
  <p>
    {% blocktrans %}
      Proctoring for this course is provided via {{ provider_name }}.  Onboarding review, including identity verification, can take 2+ business days.
    {% endblocktrans %}
  </p>
  {% if learner_notification_from_email %}
    <p>
      {% blocktrans %}
        Once your profile has been reviewed, you will receive an email with review results. The email will come from
        <a href="mailto:{{ learner_notification_from_email }}">{{ learner_notification_from_email }}</a>.
        Make sure this email has been added to your inbox filter.
      {% endblocktrans %}
    </p>
  {% endif %}
  {% if integration_specific_email %}
    <p>
      {% blocktrans %}
        Please contact
        <a href="mailto:{{ integration_specific_email }}">{{ integration_specific_email }}</a>
        if you have questions.
      {% endblocktrans %}
    </p>
  {% endif %}
{% endblock %}
{% block cta_copy %}
  {% trans "Continue to onboarding" %}
{% endblock %}

{% block cta_subtitle %}
  {% blocktrans %}
    You will be guided through online proctoring software set up and identity verification.
  {% endblocktrans %}
{% endblock %}
